/* ======================================================== */
/* ===                                                  === */
/* ===            S i n g u l a r I n p u t . c         === */
/* ===                                                  === */
/* ===	Authors: Maximilian Kreuzer, Nils-Ole Walliser	=== */
/* ===	Last update: 1/07/11                           === */
/* ===                                                  === */
/* ======================================================== */


/* ======================================================== */
/* =========            H E A D E R s             ========= */

#include "Global.h"
#include "Mori.h"

/* ======================================================== */
/* =========            D E F I N I T I O N s     ========= */

/***  local for Singularinput.c ***/
#define DijkEQ                  "->"	/* Useful for Mathematica rules: "->" */
#define	T_DIV                   "d"	    /* Toric divisors */
#define DIVclassBase            "J"	    /* Basis of the divisor classes */

/* diagnostic stuff */
#define TEST_PRINT_SINGULAR_IO  (0)     /* IO of SINGULAR */

#define NORM_SIMP_NUM           (0)

/*=========================================================*/


int Read_HyperSurf(int *he, int divclassnr, int maxline, char filename[20], MORI_Flags *_Flag){

    FILE *stream;
    
    int i;
    char string[maxline];
    char delims[] = " ";
    char *result = NULL;
	
	if(_Flag->Read_HyperSurfCounter==0){	
    		if( (stream = fopen(filename, "w")) == NULL) {
        		printf("Error: cannot open file!\n");
            		exit(1);
        	}     
 
		fgets(string, sizeof string, stdin);
		fprintf(stream, "%s\n", string);
	}
	
	if(_Flag->Read_HyperSurfCounter != 0){
		if( (stream = fopen(filename,"r")) == NULL) {
			printf("Error: cannot read file!\n");
			exit(1);	
		}
		fgets(string, maxline ,stream);
	}	

        i=0;
        result = strtok( string, delims );
        while( result != NULL ) { 
            he[i] = atoi(result); 
            i++;
            if(i == divclassnr)
                break;
            result = strtok( NULL, delims );
        }   		
        fclose(stream);

        return i;
}

void HyperSurfSingular(PolyPointList *P,triang *T, triang *SR ,MORI_Flags *_Flag , FibW *F,int *cp){
  int p=SR->v,d=SR->d,i,j,r, N=0;/* N=not(0th)=offset(IP) */
  int divclassnr = *cp;

  FILE *SF;
  char SFname[L_tmpnam], SingularCall[50+L_tmpnam], *D=T_DIV,*B=DIVclassBase;

  assert(NULL!=tmpnam(SFname));
  assert(NULL!=(SF=fopen(SFname,"w")));

  fprintf(SF,"ring r=0,(%s%d",D,N+1);	// ring r=0,(D1,...,Dp,B1,...,Bp-d),dp;
  for(i=N+2;i<=p;i++)
	  fprintf(SF,",%s%d",D,i);
  for(i=1;i<=p-d;i++)
	  fprintf(SF,",%s%d",B,i);
  fprintf(SF,"),dp;\n");

  /*LINEAR EQUIVALENCES*/
  fprintf(SF,"ideal lin=");
  for(r=0;r<d;r++){
    if(r)fprintf(SF,",");
    fprintf(SF,"0");
    for(j=0;j<p;j++){
    	int X=P->x[j][r];
    	if((X>0))
    		fprintf(SF,"+");
    	if(X)
    		fprintf(SF,"%d*%s%d",X,D,j+1);
    }
  }
  fprintf(SF,";");
  DivClassBasis(SF,P,p,D,B);

  /* STANLEY REISNER */
  fprintf(SF,"ideal sr=");
  for(r=0;r<SR->n;r++){
	  int f=1;
	  if(r)
		  fprintf(SF,",");
	  for(j=0;j<p;j++)
		  if(getN(j,SR->I[r])){
			  if(f)
				  f=0;
			  else
				  fprintf(SF,"*");
			  fprintf(SF,"%s%d",D,j+1);
		  }
  }

  /* CHOW RING */
  fprintf(SF,";\n ideal chow=std(lin+sr+DCbase);\n");

    /*reads input for the hyper surface
     * creates HEInput.txt
     * reads HEInput.txt
     * and defines poly cy as H*/
  if(_Flag->H==1){

	  if(!(_Flag->FilterFlag) && _Flag->Read_HyperSurfCounter == 0)
		  printf("Type the %d (integer) entries for the hypersurface class:\n", divclassnr);


 	int *he, i;
        he=malloc(divclassnr*sizeof(int));         
	
	for(i=0;i<divclassnr;i++)        
		he[i]=0;

	int control = Read_HyperSurf(he, divclassnr, 5*divclassnr ,"HEInput.txt", _Flag);
	
if(_Flag->Read_HyperSurfCounter==0){
         if(control != divclassnr)
                 printf("Warning: %d entries for the hypersurface class were expected\n"
                 "         (i.e. as many entries as toric divisor classes) \n",divclassnr );

fprintf(SF,"\"Hypersurface class: ");
	  for(i=0; i<divclassnr; i++)
		  if(he[i]!=0)
			  fprintf(SF,"%d*d%d ",he[i],i+1);
	  fprintf(SF,"\";\n");
}
	  fprintf(SF,"poly HySurf=(0");
	  for(i=0;i<divclassnr;i++)		 // Hypersurface equation
		  if(he[i]!=0)
			  fprintf(SF,"+%d*d%d",he[i],i+1);
	  fprintf(SF,"); \n");

	  int dim = F->nw;
	  int DegreeVec[dim];
	  for(i=0;i<F->nw;i++){
		  DegreeVec[i]=0;
	  }

	  for(i=0;i<F->nw;i++){
    		  for(j=0;j<divclassnr;j++){
    			  DegreeVec[i] += he[j]*(F->W[i][j]);
    		  }
	  }
if(_Flag->Read_HyperSurfCounter==0){
	  fprintf(outFILE,"Hypersurface degrees: (");
	  for(i=0;i<F->nw;i++){
		  fprintf(outFILE," %d ",DegreeVec[i]);
	  }
	  fprintf(outFILE,")\n");
}

	
	_Flag->Read_HyperSurfCounter++;
	
	  free(he); fflush(0);

  }

  else {
	  fprintf(SF,"poly HySurf=(%s%d",D,N+1);
	  for(i=N+1;i<p;i++)		 // CALABI-YAU
		  fprintf(SF,"+%s%d",D,i+1);
	  fprintf(SF,");\n");
  }

  fprintf(SF,"number vol=");{
	  int NN=NORM_SIMP_NUM;          // NORMALIZATION
	  Long *X[POLY_Dmax], Vijkl;
	  int x[POLY_Dmax], n=0; i=j=0;
	  if((NN<0)||(NN > T->n)) NN=0;
//  for(j=1;j<T->n;j++)if(T->I[j]<T->I[NN])NN=j; //avoid D_I.DHySurf=0 for P3xP3/Z3
    while(i<p)
    	if(getN(i++,T->I[ NN ])){
    		x[n]=i-1;
    		X[n++]=P->x[i-1];
    	}
    assert(n==P->n);
    Vijkl=SimplexVolume(X,P->n);
    fprintf(SF,"%ld;\n",Vijkl);
    fprintf(SF,"poly norm=vol*reduce(%s%d*%s%d*%s%d*%s%d,chow);\n",D,x[0]+1,D,x[1]+1,D,x[2]+1,D,x[3]+1);
  }

  if(_Flag->i || _Flag->c){
  fprintf(SF,"\"SINGULAR -> divisor classes (integral basis %s1",B);
  if(p-d>1)
	  fprintf(SF," ... %s%d",B,p-d);
  fprintf(SF,"):\";\n");
  fprintf(SF,"string LR=string(%s1)+\"=\"+string(reduce(%s1,chow));",D,D);
  for(j=2;j<=p;j++){
	  fprintf(SF,"LR=LR+\", \"+string(%s%d)+\"=\"+",D,j);
	  fprintf(SF,"string(reduce(%s%d,chow)); ",D,j);
  }
  fprintf(SF,"LR;\n");
  }

  fprintf(SF, "list dli=d1");
  for(j=2;j<=p;j++)
	  fprintf(SF,",d%d",j);
  fprintf(SF,";");

  fprintf(SF,"int i,j,k;");

  /*linear basis ideal*/
  fprintf(SF,"ideal lb=std(lin+DCbase);");


/*####### Chern classes of CY/H ################*/
	  fprintf(SF,"poly ChernUp=1;\n");
	  fprintf(SF,"for(i=1;i<=%d;i++){ChernUp=ChernUp*(1+dli[i]);}",p);
	  fprintf(SF,"poly ChernBottomHypersurface=1-(HySurf)+(HySurf)^2-(HySurf)^3;\n");

	  fprintf(SF,"poly ChernHySurf=ChernUp*ChernBottomHypersurface;\n");
	  fprintf(SF,"poly c1HySurf=jet(ChernHySurf,1)-jet(ChernHySurf,0);\n");
	  fprintf(SF,"poly c2HySurf=jet(ChernHySurf,2)-jet(ChernHySurf,1);\n");
	  fprintf(SF,"poly c3HySurf=jet(ChernHySurf,3)-jet(ChernHySurf,2);\n");


	  // PRINT HODGE and EULER
	  if(_Flag->b && _Flag->H !=1){
		  fprintf(SF,"\"SINGULAR  -> Hodge and Euler numbers of the CY:\";");
		  fprintf(SF,"\"H:\",%d,\",\",reduce(1/24*(c1HySurf*c2HySurf-12*c3HySurf)*HySurf,chow)/norm + %d,"
				  "\"[\",reduce(c3HySurf*HySurf,chow)/norm,\"]\";\n", divclassnr - P->n,divclassnr - P->n);
	  }
	  if(_Flag->b && _Flag->H ==1){
	  		  fprintf(SF,"\"SINGULAR  -> Arithmetic genera and Euler number of H:\";");
	  		  fprintf(SF,"\"chi_0:\",reduce(1/24*(c1HySurf*c2HySurf)*HySurf,chow)/norm,\",\", "
	  				  "\"chi_1:\",reduce(1/24*(c1HySurf*c2HySurf-12*c3HySurf)*HySurf,chow)/norm,"
	  				  "\" [\",reduce(c3HySurf*HySurf,chow)/norm,\"]\";\n");
	  	  }
/*#############################################*/



/*####### Intersection Polynomial #############*/
  if(_Flag->i){

	  fprintf(SF,"\"SINGULAR -> intersection polynomial:\";");
	  fprintf(SF,"poly hhh; list hli=%s1",B);
	  for(j=2;j<=p-d;j++)
		  fprintf(SF,",%s%d",B,j);
	  fprintf(SF,"; string PO=\"\";");
	  fprintf(SF,"for(i=1;i<=size(hli);i++)");
	  fprintf(SF,"{");
	  fprintf(SF,"	hhh=reduce(hli[i]^3*HySurf,chow)/norm;");
	  fprintf(SF,"	if(0!=hhh)");
	  fprintf(SF,"	{");
	  fprintf(SF,"		if(hhh>0)");
	  fprintf(SF,"		{");
	  fprintf(SF,"			PO=PO+\"+\"+string(hhh)+\"*\"+string(hli[i])+\"^3\";");
	  fprintf(SF,"		}");
	  fprintf(SF,"		else");
	  fprintf(SF,"		{");
	  fprintf(SF,"			PO=PO+string(hhh)+\"*\"+string(hli[i])+\"^3\";");
	  fprintf(SF,"		}");
	  fprintf(SF,"	}");
	  fprintf(SF,"}");
	  fprintf(SF,"for(i=1;i<=size(hli);i++)");
	  fprintf(SF,"{");
	  fprintf(SF,"	for(j=1;j<=size(hli);j++)");
	  fprintf(SF,"	{");
	  fprintf(SF,"		if(i!=j)");
	  fprintf(SF,"		{");
	  fprintf(SF,"			hhh=reduce(hli[i]*hli[i]*hli[j]*HySurf,chow)/norm;");
	  fprintf(SF,"			if(0!=hhh)");
	  fprintf(SF,"			{");
	  fprintf(SF,"				if(hhh > 0)");
	  fprintf(SF,"				{");
	  fprintf(SF,"					PO=PO+\"+\"+string(hhh)+\"*\"+string(hli[i])+\"^2*\"+string(hli[j]);");
	  fprintf(SF,"				}");
	  fprintf(SF,"				else");
	  fprintf(SF,"				{	");
	  fprintf(SF,"					PO=PO+string(hhh)+\"*\"+string(hli[i])+\"^2*\"+string(hli[j]);");
	  fprintf(SF,"				}");
	  fprintf(SF,"			} ");
	  fprintf(SF,"		}");
	  fprintf(SF,"	}");
	  fprintf(SF,"}");
	  fprintf(SF,"for(i=1;i<size(hli)-1;i++)");
	  fprintf(SF,"{");
	  fprintf(SF,"	for(j=i+1;j<=size(hli);j++)");
	  fprintf(SF,"	{");
	  fprintf(SF,"		for(k=j+1;k<=size(hli);k++)");
	  fprintf(SF,"		{");
	  fprintf(SF,"			hhh=reduce(hli[i]*hli[j]*hli[k]*HySurf,chow)/norm;");
	  fprintf(SF,"			if(0!=hhh)");
	  fprintf(SF,"			{				");
	  fprintf(SF,"				if(hhh>0)");
	  fprintf(SF,"				{");
	  fprintf(SF,"					PO=PO+\"+\"+string(hhh)+\"*\"+string(hli[i])+\"*\" +string(hli[j])+\"*\"+string(hli[k]);");
	  fprintf(SF,"				}");
	  fprintf(SF,"				else");
	  fprintf(SF,"				{");
	  fprintf(SF,"					PO=PO+string(hhh)+\"*\"+string(hli[i])+\"*\" +string(hli[j])+\"*\"+string(hli[k]);");
	  fprintf(SF,"				}");
	  fprintf(SF,"			} ");
	  fprintf(SF,"		}");
	  fprintf(SF,"	}");
	  fprintf(SF,"}");
	  fprintf(SF,"if (PO[1] == \"+\"){PO[2,size(PO)];\n}");
	  fprintf(SF,"else {PO[1,size(PO)];\n}");
 }
/*#############################################*/


  //PRINT Chern classes of CY/H
  if(_Flag->H !=1 && _Flag->c){
	  fprintf(SF,"\"SINGULAR  -> Chern classes of the CY-hypersurface:\";");
	  //fprintf(SF,"\"c1(CY)= \",reduce(c1HySurf,lb);\n");
	  fprintf(SF,"\"c2(CY)= \",reduce(c2HySurf,lb);\n");
	  fprintf(SF,"\"c3(CY)= \",reduce(c3HySurf,lb);\n");
  }
  if(_Flag->H==1 && _Flag->c){
	  fprintf(SF,"\"SINGULAR  -> Chern classes of the hypersurface H:\";");
	  fprintf(SF,"\"c1(H)= \",reduce(c1HySurf,lb);\n");
	  fprintf(SF,"\"c2(H)= \",reduce(c2HySurf,lb);\n");
	  fprintf(SF,"\"c3(H)= \",reduce(c3HySurf,lb);\n");
  }


/*####### triple intersection numbers #########*/
  if(_Flag->t){
  fprintf(SF,"poly ddd;");
  fprintf(SF,"\"SINGULAR -> triple intersection numbers:\";");
  fprintf(SF,"string SO=\"\";");
  fprintf(SF,"for(i=1;i<=%d;i++){ddd=reduce(",p);                       // di^3
  fprintf(SF,"dli[i]^3*HySurf,chow)/norm; if(0!=ddd) {SO=SO+string(dli[i])+\"^3");
  fprintf(SF,"%s\"+string(ddd)+\", \";}};SO;SO=\"\";\n",DijkEQ);

  fprintf(SF,"for(i=1;i<=%d;i++){for(j=1;j<=%d;j++)",p,p);           // di^2.dj
  fprintf(SF,"{if(i!=j){ddd=reduce(dli[i]*dli[i]*dli[j]*HySurf,chow)/norm");
  fprintf(SF,";if(0!=ddd) {SO=SO+string(dli[i])+\"^2*\"+string(dli[j])");
  fprintf(SF,"+\"%s\"+string(ddd)+\", \";} }};SO;SO=\"\";};\n",DijkEQ);

  fprintf(SF,"for(i=1;i<%d-1;i++){for(j=i+1;j<=%d;j++){",p,p);
  fprintf(SF,"for(k=j+1;k<=%d;k++){ddd=reduce(dli[i]*dli[j]*dli[k]*HySurf,chow)/norm",p);
  fprintf(SF,";if(0!=ddd) {SO=SO+string(dli[i])+\"*\"+string(dli[j])+\"*\"");
  fprintf(SF,"+string(dli[k])+\"%s\"+string(ddd)+\", \";} }}};SO;\n",DijkEQ);
}
/*#############################################*/


/*####### Top. quant'ies of the divisors '#######*/
  if(_Flag->d){
	  fprintf(SF,"\"SINGULAR -> topological quantities of the toric divisors:\";");


	  for(i=1; i<=p; i++){
		  fprintf(SF,"poly ChernBottomD%d=1-(dli[%d])+(dli[%d])^2-(dli[%d])^3;\n",i,i,i,i);
		  fprintf(SF,"poly ChernD%d=ChernUp*ChernBottomD%d*ChernBottomHypersurface;\n",i,i);
			  /*Chern classes of divisors*/
		  fprintf(SF,"poly c1D%d=jet(ChernD%d,1)-jet(ChernD%d,0);\n",i,i,i);
		  fprintf(SF,"poly c2D%d=jet(ChernD%d,2)-jet(ChernD%d,1);\n",i,i,i);
		  fprintf(SF,"poly c3D%d=jet(ChernD%d,3)-jet(ChernD%d,2);\n",i,i,i);
			  /*Todd classes as formula (5.41) in Denef lectures*/
		  fprintf(SF,"poly ToddD%d=1+1/2*c1D%d+1/12*(c1D%d^2+c2D%d)",i,i,i,i);
		  fprintf(SF,"+1/24*c1D%d*c2D%d;\n",i,i);
		  fprintf(SF,"poly Todd2D%d=jet(ToddD%d,2)-jet(ToddD%d,1);\n",i,i,i); /*=1/12(c1D^2+c2D)*/
			  /*Euler chrachteristics and arithmetic genera*/
		  fprintf(SF,"poly EulerD%d=reduce(dli[%d]*c2D%d*HySurf,chow)/norm;\n",i,i,i);
		  fprintf(SF,"poly EulerHD%d=reduce(Todd2D%d*dli[%d]*HySurf,chow)/norm;\n",i,i,i);
	  }
	  fprintf(SF,"string EulerD;");
	  fprintf(SF,"string EulerHD;");
	  for(i=1;i<=p;i++){
		  fprintf(SF,"EulerD=EulerD+string(EulerD%d)+\" \";",i);
		  fprintf(SF,"EulerHD=EulerHD+string(EulerHD%d)+\" \";",i);
	  }
		  fprintf(SF,"\"Euler characteristics:\", EulerD;\n ");
		  fprintf(SF,"\"Arithmetic genera:\", EulerHD;\n");



		  /*Singular lists out of the top. quant'ies*/
		  /*This makes live easier*/
	  fprintf(SF,"list c1D; list eulerHD; list eulerD;");
	  for(i=1;i<=p;i++){
		  fprintf(SF,"c1D[%d]=c1D%d;",i,i);
		  fprintf(SF,"eulerHD[%d]=EulerHD%d;",i,i);
		  fprintf(SF,"eulerD[%d]=EulerD%d;",i,i);
	  }
	  fprintf(SF,";\n");
/*#############################################*/

/*####### dP conditions #######################*/
	  fprintf(SF,"kill k;");
	  fprintf(SF,"list dP; list N; int l=0; int p=0; int t=0; int s=0; int k=0; int u=0; int v=0;");
	 	  fprintf(SF,""
	 			  "for(i=1;i<=size(dli);i++){"
	 				  "if(eulerHD[i]==1 && eulerD[i]>=3 && eulerD[i]<=11 && reduce(c1D[i]^2*dli[i]*HySurf,chow)/norm >=1 && reduce(c1D[i]^2*dli[i]*HySurf,chow)/norm <= 9){"
	 					  "for(j=1;j<=size(dli);j++){"
	 							  "if(dli[j] != dli[i]){"
									  "for(k=1;k<=size(dli);k++){"
										  "v++;"
										  "if(reduce(dli[i]*dli[j]*dli[k]*HySurf,chow)/norm == 0){u++;}"
									  "}"
									  "if(u < v){"
	 								  "if(reduce(dli[j]*dli[i]*c1D[i]*HySurf,chow)/norm <= 0){"
	 									  "p++;"
	 								  "}"
									  "} u=0; v=0;"
	 							  "}"
	 						  "}"
	 					  "if(p==0){"
	 							  "s++;"
	 						  "dP[s]=dli[i];"
	 						  "N[s]=eulerD[i]-3;"
	 					  "}"
	 				  "}"
	 			  "p=0;"
	 			  "t=0;"	/* index for no repetitions of div. classes */
	 			  "}	");

	 	  fprintf(SF,"string DelPezzo;for(i=1;i<=size(dP);i++){DelPezzo=DelPezzo+string(dP[i])+\"(\"+string(N[i])+\")\"+\" \";};");
	 	  fprintf(SF,"list nonintDP; int g=0; int f=0; if(size(dP)>0){if(size(dP)>1){for(i=1;i<=size(dP);i++){for(j=1;j<=size(dP);j++){if(dP[j]!=dP[i]){for(k=1;k<=size(dli);k++){if(reduce(dP[i]*dP[j]*dli[k]*HySurf,chow)/norm!=0){g++;} ;};};}; if(g==0){f++; nonintDP[f]=dP[i];}; g=0;};} else {nonintDP[1]=dP[1];};};");
	 	  fprintf(SF,"string niDP;for(i=1;i<=size(nonintDP);i++){niDP=niDP+string(nonintDP[i])+\" \";};");
	 	  fprintf(SF, "\"dPs:\",size(dP),\";\",DelPezzo,\"nonint:\",size(nonintDP),\";\",niDP;");
/*#############################################*/
}

  fprintf(SF,"quit;\n"); fflush(0); fclose(SF);

#if (TEST_PRINT_SINGULAR_IO)
				CatFile(SFname);
#endif

  strcpy(SingularCall,"Singular -q < "); strcat(SingularCall,SFname);
  assert(strlen(SingularCall)<50+L_tmpnam);
  if( system(SingularCall) ) {puts("Check Singular installation");exit(1);}
  remove(SFname);}



