/* =========================================================== */
/* ===                                                     === */
/* ===                  m o r i . c                        === */
/* ===                                                     === */
/* ===	Authors: Maximilian Kreuzer, Nils-Ole Walliser	   === */
/* ===	Last update: 22/06/11                              === */
/* ===                                                     === */
/* =========================================================== */	


/* ======================================================== */
/* =========            H E A D E R s             ========= */

#include "Global.h"
#include "LG.h"
#include "Mori.h"

/*==========================================================*/

FILE *inFILE, *outFILE;

void  PrintUsage(char *c){
  printf("This is ``mori.x'':  performing crepant star triangulations of a polytope P* in the N-lattice\n");
  printf("                     computing the Mori cone of the corresponding toric ambient spaces\n");
  printf("                     determining the intersection rings of embedded (CY) hypersurfaces\n");
  printf("Usage:   %s [-<Option-string>] [in-file [out-file]]\n", c);
  printf("Options (concatenate any number of them into <Option-string>):\n");

  printf("    -h      print this information \n");
  printf("    -f      use as filter\n");
  printf("    -g      general output: triangulation and Stanley-Reisner ideal\n");
  printf("    -I      incidence information of the facets (ignoring interior points of facets)\n");
  printf("    -m      Mori generators of the ambient space\n");
  printf("    -P      IP-simplices among points of P* (ignoring interior points of facets)\n");
  printf("    -K      points of P* in Kreuzer polynomial form\n");
  printf("    -b      Hodge numbers and Euler number\n");
  printf("            or arithmetic genera and Euler number if combined with -H\n");
  printf("    -i      intersection ring\n");
  printf("    -c      Chern classes of the (CY) hypersurface\n");
  printf("    -t      triple intersection numbers\n");
  printf("    -d      topological information on toric divisors and del Pezzo conditions\n");
  printf("    -a      all of the above except h, f, I and K\n");
  printf("    -D      lattice polytope points of P* as input (default CWS)\n");
  printf("    -H      arbitrary hypersurface class `H = c1*D1 + c2*D2 + ...' as input (default CY)\n");
  printf("            Input: coefficients `c1 c2 ...'\n");
  puts("Input: 1) standard input format: degrees and weights `d1 w11 w12 ... d2 w21 w22 ...'");
  puts("       2) alternative (use -D): `d np' or `np d' (d=Dimension, np=#[points])");
  puts("                                and (after newline) np*d coordinates");
  puts("Output:   as specified by options");
}

int main (int narg, char* fn[]){


  int  n=0, k,
       IP,// =1 if IP property 
       R; // =1 if all poly facets at distance 1 from origin

  /* flags */
  MORI_Flags Flag;
  
       Flag.FilterFlag = 0; // filter
       Flag.g = 0; // -g: general output
       Flag.m = 0; // -m: Mori cone
       Flag.P = 0; // -P: IP simplices
       Flag.K = 0; // -K: Newton polynomial
       Flag.i = 0; // -i: intersection ring
       Flag.t = 0; // -t: triple intersection number
       Flag.c = 0; // -c: Chern classes
       Flag.d = 0; // -d; del Pezzo
       Flag.a = 0; // -a: all of the above except h,f and K
       Flag.b = 0; // -b: Hodge number of toric div
       Flag.D = 0; // -D: dual poly as input
       Flag.H = 0; // -H: arbitrary hypersurface
       Flag.I = 0; // -I: incidence information
       Flag.Read_HyperSurfCounter = 0; // see Mori.h for description
  char c;

  CWS *CW=(CWS *) malloc(sizeof(CWS));

  VertexNumList V;
  EqList *E = (EqList *) malloc(sizeof(EqList));
  EqList *DE = (EqList *) malloc(sizeof(EqList));
  
  PolyPointList *_P = (PolyPointList *) malloc(sizeof(PolyPointList)),
               *_DP = (PolyPointList *) malloc(sizeof(PolyPointList));

  PairMat *PM = (PairMat *) malloc(sizeof(PairMat)),
         *DPM = (PairMat *) malloc(sizeof(PairMat));

  if((CW==NULL)||(E==NULL)||(_P==NULL)||(DE==NULL)||(_DP==NULL)||(PM==NULL)||(DPM==NULL)){
    puts("Allocation failure: Reduce dimensions!");
    exit(0);
  }
  CW->nw=0;

  while(narg > ++n) {
    if(fn[n][0]!='-') break;
    k=0;
    while ((c=fn[n][++k])!='\0'){
      if(c=='h') { PrintUsage(fn[0]); exit(0);}
      if(c=='f') Flag.FilterFlag=1;
      if(c=='g') Flag.g=1;
      if(c=='m') Flag.m=1;
      if(c=='P') Flag.P=1;
      if(c=='K') Flag.K=1;
      if(c=='i') Flag.i=1;
      if(c=='t') Flag.t=1;
      if(c=='c') Flag.c=1;
      if(c=='d') Flag.d=1;
      if(c=='a') Flag.a=1;
      if(c=='b') Flag.b=1;
      if(c=='D') Flag.D=1;
      if(c=='H') Flag.H=1;
      if(c=='I') Flag.I=1; 
      }
  }
  n--;

  if(Flag.g + Flag.m + Flag.P + Flag.K + Flag.i + Flag.t + Flag.c + Flag.d
		  + Flag.a + Flag.b + Flag.H
		  + Flag.I ==0) Flag.g=1;

  if(Flag.a){
	  Flag.g=1;
	  Flag.m=1;
	  Flag.P=1;
	  //Flag.K=1;
	  Flag.i=1;
	  Flag.t=1;
	  Flag.c=1;
	  Flag.d=1;
	  Flag.b=1;
  }

  if(Flag.H==1 && (Flag.g + Flag.m + Flag.P + Flag.K + Flag.i + Flag.t + Flag.c + 
           Flag.d + Flag.a + Flag.b + Flag.I ==0 )){
	Flag.b=1;
	//Flag.g=1;	
	}

  if(Flag.FilterFlag) {inFILE=NULL; outFILE=stdout;}

  else {
    if (narg > ++n)  inFILE=fopen(fn[n],"r");
    else inFILE=stdin;

    if (inFILE==NULL){printf("Input file %s not found!\n",fn[n]);exit(0);}

    if (narg > ++n) outFILE=fopen(fn[n],"w");
    else outFILE=stdout;
  }

  /*=================   If input is CWS  ======================*/

  if(Flag.D ==0)
  while(Read_CWS(CW,_P)) {
    R=0;
    /* if input is IP */
    if ((IP=Find_Equations(_P,&V,E))){
    	if(CW->nw == 0){puts("Error: expected input format is CWS!"); break;}
    	R=EL_to_PPL(E, _DP, &_P->n);
    }
    if (!R) {
    	fprintf(outFILE,"Input not reflexive!\n");
    	continue;
    }

    Sort_VL(&V);
    Make_VEPM(_P,&V,E,*PM);

    if(Flag.D || !(CW->nw)) Complete_Poly(*PM,E,V.nv,_P);
    if(R&&!(Flag.D &&(CW->nw))) {
    	if(0==Transpose_PM(*PM, *DPM, V.nv, E->ne)){
    		fprintf(stderr,"Transpose_PM failed because #eq=%d > VERT_Nmax\n",E->ne);
    		exit(0);
    	}
    	VNL_to_DEL(_P, &V, DE);
    	Complete_Poly(*DPM,DE,E->ne,_DP);
    }

    if(R)
      {
    	int nv=(!(Flag.D))*E->ne;
    	int i;

    	FibW *F = (FibW *) malloc( sizeof(FibW) );

    	assert(F!=NULL);

    	for (i=nv; i<_DP->np-1; i++)
    		if(Vec_is_zero(_DP->x[i],_DP->n)) {
    			Swap_Vecs(_DP->x[i],_DP->x[_DP->np-1],_DP->n);
    			break;
    		}
    	EqList E;
   		assert(Ref_Check(_DP,&V,&E));
   		HyperSurfDivisorsQ(_DP,&V,&E,F, &Flag);

   		free(F);
    }


    fflush(outFILE);
  }



  /*=================   If input if PP  ======================*/
  if(Flag.D)
  while(Read_PP(CW,_P)) {
    R=0;
    /* if input is IP */
    if ((IP=Find_Equations(_P,&V,E))){ 

    		if(CW->nw != 0){ puts("Error: expected input format is matrix of polytope points!"); break;}
    		int k;
    		*_DP=*_P;
    		R=EL_to_PPL(E, _P, &_P->n);
    		VNL_to_DEL(_DP, &V, E);
    		for (k=0;k<_P->np;k++) V.v[k]=k;
    		V.nv=_P->np;
    }
    if (!R) {
    	fprintf(outFILE,"Input not reflexive!\n");
    	continue;
    }

    Sort_VL(&V);
    Make_VEPM(_P,&V,E,*PM);

    if(Flag.D || !(CW->nw)) Complete_Poly(*PM,E,V.nv,_P);
    if(R&&!(Flag.D &&(CW->nw))) {
    	if(0==Transpose_PM(*PM, *DPM, V.nv, E->ne)){
    		fprintf(stderr,"Transpose_PM failed because #eq=%d > VERT_Nmax\n",E->ne);
    		exit(0);
    	}
    	VNL_to_DEL(_P, &V, DE);
    	Complete_Poly(*DPM,DE,E->ne,_DP);
    }

    if(R)   
      {
    	int nv=(!(Flag.D))*E->ne;
    	int i;

    	FibW *F = (FibW *) malloc( sizeof(FibW) );

    	assert(F!=NULL);

    	for (i=nv; i<_DP->np-1; i++)
    		if(Vec_is_zero(_DP->x[i],_DP->n)) {
    			Swap_Vecs(_DP->x[i],_DP->x[_DP->np-1],_DP->n);
    			break;
    		}
    	EqList E;
   		assert(Ref_Check(_DP,&V,&E));
   		HyperSurfDivisorsQ(_DP,&V,&E,F, &Flag); 
   		
   		free(F);
    }
    
    fflush(outFILE);
  }


  return 0;
}

